;;; rcd-db-basics.el --- RCD Database Basics  -*- lexical-binding: t; -*-

;; Copyright (C) 2016-2024 by Jean Louis

;; Author: Jean Louis <bugs@gnu.support>
;; Version: 1.44
;; Package-Requires: ((rcd-utilities "1.90"))
;; Keywords: applications
;; URL: https://hyperscope.link/3/7/4/9/3/RCD-Database-Basics-37493.html

;; This file is not part of GNU Emacs.

;; This program is free software: you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation, either version 3 of the
;; License, or (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
;; General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program. If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; RCD is acronym for Reach, Connect, Deliver, my personal
;; principle and formula for Wealth.
;;
;; RCD Database Basics deals with general functions for any database
;; type used within RCD Notes for GNU Emacs.
;;
;; RCD Notes for GNU Emacs has its core in the database design
;; principles as dictated by GeDaFe:
;;
;; GeDaFe - PostgreSQL Generic Database Interface
;; http://gedafe.github.io/doc/gedafe-sql.en.html

;; Thus for the RCD Notes we always have the column `TABLENAME_id' in
;; every table as it follows the GeDaFe database design. This column
;; must be PRIMARY KEY.
;;
;; RCD Notes database design introduces requirement for every table to
;; have following columns:
;; 
;; "TABLENAME_datecreated"
;; "TABLENAME_datemodified"
;; "TABLENAME_usercreated"  -- if database support users
;; "TABLENAME_usermodified -- if database support users
;; "TABLENAME_name NOT NULL DEFAULT '>>>PLACEHOLDER<<<'"
;; or "TABLENAME_name" may be "GENERATED ALWAYS" column
;; "TABLENAME_description"

;;; Change Log:
 
;;; Code:

;;; REQUIRE SECTION

(require 'rcd-utilities)

;;; VARIABLES

(defvar rcd-db-sql-logging nil
  "Log SQL queries when true.")

(defvar rcd-db-sql-message-function 'rcd-message
  "The RCD SQL message function, like the function `message-box'")

(defvar rcd-db-columns-to-read-color '(
				"hyobjectypes_emacsforegroundcolor"
				"hyobjectypes_emacsbackgroundcolor"
				)
  "Columns to read Emacs color.")

(defvar rcd-db-columns-to-read-date-time '(
				"messages_scheduled"
				)
  "Columns to read date time.")

(defvar rcd-db-people-prompt "Search for people: ")

;;;; DATABASE VARIABLES

(defvar rcd-tabulated-entry-limit 1000
  "Limit  entries in `tabulated-list-mode'.")


;;; RCD Notes functions

(defun rcd-notes-name (&optional text-1 text-2 text-3)
  "Return RCD Notes program name with optionally concatenated text."
  (cond ((and text-1 text-2 text-3) (format "%s → %s → %s → %s" rcd-program-name text-1 text-2 text-3))
	((and text-1 text-2) (format "%s → %s → %s" rcd-program-name text-1 text-2))
	(text-1 (format "%s → %s" rcd-program-name text-1))
	(t rcd-program-name)))

;;; CUSTOM VARIABLES

(defcustom rcd-db-people-related-org-property "PEOPLE-DB-ID"
  "Related Org property for people ID. 

Property shall accept the database type, hostname or file name,
port, table and ID, such as in the example:

:PEOPLE-DB-ID: pg:localhost:5432:people:1
:PEOPLE-DB-ID: pg:::people:1
:PEOPLE-DB-ID: sqlite:::people:1
:PEOPLE-DB-ID: sqlite:~/.emacs.d/rcd-people.sqlit::people:1"
  :type 'string
  :group 'rcd-db)

;;; DB MODE MAP

(defvar rcd-db-mode-map
  (let ((map (make-sparse-keymap)))
    (set-keymap-parent map tabulated-list-mode-map)
    (keymap-set map "/" #'rcd-tabulated-filter)
    (keymap-set map "g" #'goto-line)
    (keymap-set map "q" #'(lambda () (interactive) (rcd-notes-kill-buffers)))
    (keymap-set map "u" #'rcd-tabulated-remove-marks)
    (keymap-set map "v" #'rcd-tabulated-id-to-register)
    (keymap-set map "\\" #'rcd-tabulated-filter-reset)
    (keymap-set map "a" #'rcd-db-insert-new-row)
    (keymap-set map "d" #'rcd-db-delete-tab-database-entry)
    (keymap-set map "e" #'rcd-db-tabulated-edit-database-entry)
    (keymap-set map "g" #'rcd-tabulated-refresh)
    (keymap-set map "j" #'next-line)
    (keymap-set map "k" #'previous-line)
    (keymap-set map "m" #'rcd-tabulated-mark-id)
    (keymap-set map "q" #'rcd-db-quit-window)
    (keymap-set map "u" #'rcd-tabulated-unmark-id)
    (keymap-set map "U" #'rcd-tabulated-remove-marks)
    map)
  "The basic RCD database keymap.")

(defun rcd-db-table-mode-map (table)
  "Generate mode map for table.

Use `rcd-db-current-database-type'."
  (let ((symbol (format "rcd-db-%s-%s-mode-map" rcd-db-current-database-type table)))
    (if (intern-soft symbol)
	(symbol-value (intern symbol))
      rcd-db-mode-map)))

;;; GENERIC RCD DATABASE MODE MAP

(defun rcd-db-tabulated-edit-database-entry ()
  "Edit tabulated database entry."
  (interactive)
  (let ((id (tabulated-list-get-id)))
    ;;(refresh rcd-tabulated-refresh-function)
    ;; (point (point)))
    (when id
      (rcd-db-tabulated-edit-entry rcd-db-current-table id)))) ;;rcd-db-edited-table)))) todo fix

(defun rcd-db-tabulated-edit-entry (table &optional id)
  (interactive)
  (when-tabulated-id 'any
      (let* ((database-type (or rcd-db-current-database-type "pg")) ;; TODO this is not correct to inject pg
	     (symbol (format "rcd-db-%s-tabulated-table-edit" database-type)))
	(unless database-type
	  (error "Can't find database type from `rcd-db-current-database-type'"))
	;; (unless rcd-db-current-database-handle
	;;   (error "Can't find current database handle from `rcd-db-current-database-handle'"))
	(cond ((intern-soft symbol) (funcall (intern symbol) table id))
	      (t (error "Can't invoke `%s'" symbol))))))

;;; WINDOW FUNCTIONS

(defun rcd-db-quit-window (&optional kill)
  "Quit database window."
  (interactive)
  (cond (rcd-current-return-function (progn
				       (quit-window kill)
				       (funcall rcd-current-return-function)))
	(t (quit-window kill))))

;;; DB INSERT NEW ROW

(defun rcd-db-insert-new-row (&optional table)
  "Generic function to insert new row."
  (interactive)
  (cond ((or (and table rcd-db-current-database-type)
	     (and rcd-db-current-table rcd-db-current-database-type))
	 (let* ((table (or table rcd-db-current-table))
		(function-name (format "rcd-db-table-%s-%s-insert-new-row" rcd-db-current-database-type table))
		(symbol (intern-soft function-name)))
	   (if symbol
	       (funcall symbol)
	     (rcd-warning-message "Cannot insert new row for table `%s' as function `%s' does not exist" table function-name))))
	(t (rcd-warning-message "Not in RCD database environment"))))

;;; DB REPORT FUNCTIONS

(defun rcd-db-report (title entries format database-type db-handle table sort-key 
			    &optional refresh highlight-list place id return-function)
  "RCD database report.

TITLE is buffer name."
  (setq rcd-db-current-database-type database-type)
  (let* ((buffer (generate-new-buffer-name (rcd-notes-name title))))
    (let* ((buffer (get-buffer-create buffer))
	   (mode-map (rcd-db-table-mode-map table)))
      (cond ((string-equal table "edit") (switch-to-buffer buffer))
	    (rcd-dont-switch-windows (switch-to-buffer buffer))
	    (t (switch-to-buffer-other-window buffer)))
      (setq tabulated-list-format format)
      (setq tabulated-list-entries entries)
      (setq rcd-db-edited-table table)
      (setq rcd-db-current-database-type database-type)
      (setq rcd-tabulated-refresh-function refresh)
      (setq rcd-current-return-function return-function)
      (rcd-db-list-mode)
      (use-local-map mode-map)
      (setq rcd-tabulated-marked-items nil)
      (setq rcd-db-current-database-handle db-handle)
      (setq rcd-db-current-table (or (alist-get "table" place nil nil 'equal) table))
      (setq rcd-db-current-table-id id)
      (setq tabulated-list-padding 1)
      (tabulated-list-init-header))
    (setq tabulated-list-sort-key sort-key)
    (tabulated-list-print t)
    (when highlight-list
      (rcd-highlight-list highlight-list))))

;;; SQL 

(defun rcd-sql-hash-with-key (sql pg)
  "Return hash with key TEXT [ID] and value ID from SQL result.

Output is in the form ID, TEXT
Argument PG is database handle."
  (let ((hash (make-hash-table :test 'equal))
	(res (rcd-sql sql pg)))
    (while res
      (let ((item (pop res)))
	(cond ((eq (type-of item) 'vector) (puthash (format "%s [%s]" (elt item 1) (elt item 0)) (elt item 0) hash)))))
    hash))

(defun rcd-sql-hash (sql pg)
  "Return hash with by using SQL and PG database handle.

Output is in the form TEXT, ID"
  (let ((hash (make-hash-table))
	(list (rcd-sql-list sql pg)))
    (while list
      (let ((item (pop list)))
	(puthash (format "%s" (elt item 0)) (elt item 1) hash)))
    hash))

;;; Primary Key

;;; SQL QUERY CONTROL

(defun rcd-sql-begin (pg)
  "Send BEGIN query to database by using PG database handle."
  (rcd-sql "BEGIN" pg))

(defun rcd-sql-commit (pg)
  "Send COMMIT query to database by using PG database handle."
  (rcd-sql "COMMIT" pg))

(defun rcd-sql-rollback (pg)
  "Send ROLLBACK query to database by using PG database handle."
  (rcd-sql "ROLLBACK" pg))

;;; COMPLETING READ WITH SQL

(defun rcd-completing-id-by-sql (prompt sql pg)
  "Return the ID by using SQL query.

PROMPT will be displayed to user.
PG is database handle

This function will provide entries with the ID number at
beginning, and it will cut the number from string and return it.

The SQL query has to give just one column with the number
separate by space with other column values.

Example would be: SELECT concat_ws(' ',1,'My ID') that will return `1'"
  (let* ((selection (rcd-sql-list sql pg))
	 (choice (completing-read prompt selection)))
    (string-cut-id choice)))

(defun rcd-completing-id-by-sql-end (prompt sql pg)
  "Return the ID as by finding [ID:123] on the end.
Argument PROMPT will be displayed to user.
Argument SQL is supplied query.
Argument PG is database handle."
  (let* ((selection (rcd-sql-list sql pg))
	 (choice (completing-read prompt selection)))
    (rcd-get-bracketed-id-end choice)))

(defun rcd-completing-read-sql-hash (prompt sql pg &optional history initial-input not-require-match auto-initial-input)
  "Complete selection by using SQL.

First column shall be unique id, followed by text
representation.  Example SQL query:

SELECT people_id, people_firstname || ' ' || people_lastname FROM people

PG is database handle.  HISTORY is supported with INITIAL-INPUT
Argument PROMPT will be displayed to user."
  (let* ((gc-cons-threshold 5000000)
	 (hash (rcd-sql-hash-with-key sql pg))
	 (completion-ignore-case t)
	 (require-match (if not-require-match nil t))
	 (history (or history (rcd-ask-history-variable prompt)))
	 (initial-input (or initial-input (when auto-initial-input (car (symbol-value history)))))
	 (choice (completing-read prompt hash nil require-match initial-input history))
	 (choice (string-trim choice))
	 (id (gethash choice hash)))
    (if id id
      (if not-require-match 
	  choice))))

;;; DATABASE INFORMATION

(defun rcd-db-current-database (pg)
  "Return database name for handle PG."
  (rcd-sql-first "SELECT current_database()" pg))

(defun rcd-db-table-exists-p  (table pg &optional schema)
  "Return T if TABLE exists for handle PG.

SCHEMA is optional.

Views are considered TABLE."
  (let* ((schema (or schema "public"))
	 (exists (rcd-sql-first 
		  "SELECT EXISTS 
                 (SELECT 1 FROM information_schema.tables 
                  WHERE table_schema = $1 
                  AND table_name = $2)"
		  pg schema table)))
    (cond (exists table)
	  (nil nil))))

(defun rcd-db-tables-list (pg)
  "Return list of all tables for handle PG."
  (rcd-sql-list "SELECT c.relname FROM pg_catalog.pg_class c LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace WHERE c.relkind IN ('r','p','') AND n.nspname <> 'pg_catalog' AND n.nspname <> 'information_schema' AND n.nspname !~ '^pg_toast' AND pg_catalog.pg_table_is_visible(c.oid) ORDER BY c.relname" pg))

(defun rcd-db-table-full-attributes (oid pg)
  "Return full attributes for OID with database handle PG."
  (rcd-sql "SELECT a.attname, pg_catalog.format_type(a.atttypid, a.atttypmod), (SELECT substring(pg_catalog.pg_get_expr(d.adbin, d.adrelid) for 128) FROM pg_catalog.pg_attrdef d WHERE d.adrelid = a.attrelid AND d.adnum = a.attnum AND a.atthasdef), a.attnotnull, a.attnum, (SELECT c.collname FROM pg_catalog.pg_collation c, pg_catalog.pg_type t WHERE c.oid = a.attcollation AND t.oid = a.atttypid AND a.attcollation <> t.typcollation) AS attcollation, a.attidentity, NULL AS indexdef, NULL AS attfdwoptions, a.attstorage, CASE WHEN a.attstattarget=-1 THEN NULL ELSE a.attstattarget END AS attstattarget, pg_catalog.col_description(a.attrelid, a.attnum) FROM pg_catalog.pg_attribute a WHERE a.attrelid = $1 AND a.attnum > 0 AND NOT a.attisdropped ORDER BY a.attnum" 
	   pg oid))

(defun rcd-db-current-database-name (pg)
  "Return current database name for handle PG."
  (rcd-sql-first "SELECT description FROM pg_shdescription 
                  JOIN pg_database ON objoid = pg_database.oid 
                  WHERE datname = $1"
		 pg (rcd-db-current-database pg)))

(defun rcd-db-table-column-referenced-by-tables (table column pg &optional schema database)
  "Return list of tables referencing TABLE and COLUMN.

PG is database handle."
  (rcd-sql-list "SELECT DISTINCT r.table_name
  FROM information_schema.constraint_column_usage u
       INNER JOIN information_schema.referential_constraints fk
           ON u.constraint_catalog = fk.unique_constraint_catalog
           AND u.constraint_schema = fk.unique_constraint_schema
           AND u.constraint_name = fk.unique_constraint_name
       INNER JOIN information_schema.key_column_usage r
           ON r.constraint_catalog = fk.constraint_catalog
           AND r.constraint_schema = fk.constraint_schema
           AND r.constraint_name = fk.constraint_name
 WHERE
    u.table_catalog = $1 AND
    u.table_name = $2 AND
    u.column_name = $3 AND
    u.table_schema = $4"
		pg (or database (rcd-db-current-database pg))
		table column (or schema "public")))

(defun rcd-db-table-oid (table pg &optional schema)
  "Return the TABLE OID for database handle PG.

Use optional SCHEMA."
  (let* ((schema (or schema "public"))
	 (sql (format "SELECT '%s.%s'::regclass::oid" schema table)))
    (rcd-sql-first sql pg)))

(defun rcd-db-table-comment-by-oid (oid pg)
  "Return table comment by OID.

PG is database handle."
  (rcd-sql-first "SELECT pg_catalog.obj_description($1)" pg oid))

(defun rcd-db-table-update-comment (table comment pg)
  "Update TABLE with COMMENT for handle PG."
  (rcd-sql (format "COMMENT ON TABLE %s IS %s" table (sql-escape-string comment)) pg))

(defun rcd-db-table-name (table pg)
  "Return TABLE name."
  (rcd-sql-first "SELECT description FROM pg_description JOIN pg_class ON pg_description.objoid = pg_class.oid WHERE relname = $1" pg table))

(defun rcd-db-table-by-oid (oid pg)
  "Return table name for OID using database handle PG."
  (rcd-sql-first (format "SELECT %d::regclass" oid) pg))

(defun rcd-db-table-hash (table id pg)
  "Return hash with TABLE column names and values for ID and PG."
  (let* ((oid (rcd-db-table-oid table pg))
	 (attributes (rcd-db-table-full-attributes oid pg))
	 (hash (make-hash-table :test #'equal)))
    (while attributes
      (let* ((attr (pop attributes))
	     (column (aref attr 0))
	     (type (aref attr 1))
	     (comment (aref attr 11))
	     (value (rcd-db-get-entry table column id pg)))
	(puthash (format "%s:%s:%s" table column id) ;; provide attributes to ID
		 (list comment type value)
		 hash)))
    hash))

(defun rcd-db-table-columns-list (table pg)
  "Return list of columns for TABLE and connection PG."
  (let* ((oid (rcd-db-table-oid table pg))
	 (attributes (rcd-db-table-full-attributes oid pg)))
    (mapcar (lambda (column) (aref column 0)) attributes)))

(defun rcd-db-column-type (table column pg)
  "Return type of the COLUMN in the table TABLE.
PG is database handle."
  (rcd-sql-first (format "SELECT data_type FROM information_schema.columns WHERE table_name = '%s' AND column_name = '%s'" table column) pg))

(defun rcd-db-column-foreign-table (table column pg)
  "Return the foreign table referenced by COLUMN in the TABLE.
PG is database handle."
  (rcd-sql-first (format "SELECT a.confrelid::regclass FROM pg_constraint a JOIN pg_attribute b
                          ON a.conrelid = b.attrelid
                          AND b.attnum = any (a.conkey)
                          WHERE a.conrelid = '%s'::regclass
                          AND a.contype = 'f'
                          AND b.attname = '%s'" table column) pg))

(defun rcd-db-comment (database pg)
  "Returns the database comment"
  (rcd-sql-first
   "SELECT description 
    FROM pg_shdescription 
    JOIN pg_database ON objoid = pg_database.oid 
    WHERE datname = $1" 
   pg database))

(defun rcd-db-column-comment (table column pg &optional schema)
  "Return comment for COLUMN within TABLE by using handle PG.

SCHEMA is optional and by default `public'."
  (let ((schema (or schema "public")))
    (rcd-sql-first 
     "SELECT pgd.description 
      FROM pg_catalog.pg_statio_all_tables AS st 
      INNER JOIN pg_catalog.pg_description pgd ON (pgd.objoid=st.relid) 
      INNER JOIN information_schema.columns c ON (pgd.objsubid=c.ordinal_position 
       AND c.table_schema=st.schemaname 
       AND c.table_name=st.relname 
       AND c.table_name = $1 
       AND c.table_schema = $2 
       AND c.column_name = $3)" 
     pg table schema column)))

(defun rcd-db-table-last-id (table pg)
  "Return the last entry ID for the TABLE with database handle PG."
  (rcd-sql-first (format "SELECT %s_id FROM %s ORDER BY %s_id DESC LIMIT 1" table table table) pg))

;;; INDEXES

(defun rcd-db-indexes-list (pg &optional schema)
  "Return list of all indexes.

SCHEMA is optional, by default \"public\"."
  (let ((schema (or schema "public")))
    (rcd-sql-list "SELECT indexname FROM pg_indexes WHERE schemaname = $1 ORDER BY indexname" pg schema)))

;;; DATABASE MANAGEMENT

(defun rcd-db-get-entry (table column id pg)
  "Return value for the COLUMN from the RCD Database by using the entrxy ID.
Argument TABLE is database table."
  (if id
      (let* ((sql (format "SELECT %s FROM %s WHERE %s_id = %s" column table table id))
	     (value (rcd-sql-first sql pg)))
      value)
    (error "Function `rcd-db-get-entry': ID not conveyed")))

(defun rcd-db-get-entry-auto (table column id pg)
  "Return value for TABLE, COLUMN by using ID and database handle PG.

This function finds primary key automatically."
  (cond (id (let* ((primary-key (elt (rcd-db-table-primary-key table pg) 0))
		   (sql (format "SELECT %s FROM %s WHERE %s = %s" column table primary-key id)))
	      (rcd-sql-first sql pg)))
	(t (error "ID not conveyed"))))

(defun rcd-db-get-entry-array-string (table array-column id pg &optional separator null-string-text)
  "Return value for the ARRAY-COLUMN from the RCD Database by using the entry ID.
Argument TABLE is database table."
  (if id
      (let* ((separator (or separator ","))
	     (null-string-text (or null-string-text "*"))
	     (array-column (format " array_to_string(%s, '%s', '%s')" array-column separator null-string-text))
	     (sql (format "SELECT %s FROM %s WHERE %s_id = %s" array-column table table id))
	     (value (rcd-sql-first sql pg)))
      value)
    (error "ID not conveyed")))

(defun rcd-db-get-entry-array-list (table array-column id pg)
  "Return list for the ARRAY-COLUMN from the RCD Database by using the entry ID.
Argument TABLE is database table."
  (if id
      (let* ((separator "␞")
	     (null-string-text "")
	     (value (rcd-db-get-entry-array-string table array-column id pg separator null-string-text)))
	(when value
	  (split-string value separator t " ")))
    (error "ID not conveyed")))

(defun rcd-db-string-array-to-list (string-array &optional type)
  "Return list after parsing STRING-ARRAY within curly brackets `{}'

STRING-ARRAY is in format {something}

Optional TYPE can be `integer'"
  (if string-array
      (let ((string-array (string-trim string-array)))
	(if (and (string-match "^{" string-array)
		 (string-match "}$" string-array))
	    (let* ((string-array (substring string-array 1 (1- (length string-array))))
		   (list (split-string string-array "," nil " "))
		   (type (or type 'string))
		   (list (if (eq type 'integer) (mapcar 'string-is-number-p list) list)))
	      list)
	  (rcd-warning-message "String array `%s' seem not to be array" string-array)))
    '()))

(defun rcd-db-get-entry-where (table column where pg)
  "Return value from database COLUMN in TABLE by using WHERE.
PG is database handle."
  (let ((sql (format "SELECT %s FROM %s WHERE %s" column table where)))
    (rcd-sql-first sql pg)))

;;; TABLE VIEWS

(defun rcd-db-views-list (pg &optional schema)
  "Return list of views for database handle PG.

SCHEMA is optional and by default \"public\"."
  (rcd-sql-list "SELECT table_name FROM information_schema.views
                 WHERE table_schema NOT IN ('information_schema', 'pg_catalog')
                 AND table_schema = $1
                 ORDER BY table_name"
		pg (or schema "public")))

(defun rcd-db-views-select (pg &optional schema)
  "Return selected table view for database handle PG.

SCHEMA is optional and by default \"public\"."
  (rcd-choose (rcd-db-views-list pg schema) "Choose table view: "))

;;; ENCRYPTION AND DECRYPTION

(defun rcd-db-get-encrypted-entry (table column id symmetric pg)
  "Return decrypted value for TABLE and COLUMN with its ID.
SYMMETRIC is provided password.
Argument PG is database handle."
  (let ((sql (format "SELECT pgp_sym_decrypt(%s::bytea,'%s') FROM %s WHERE %s_id = %d" column symmetric table table id)))
    (rcd-sql-first sql pg)))

(defun rcd-db-update-encrypted-entry (table column value id symmetric pg)
  "Update TABLE, COLUMN by its ID, with VALUE and SYMMETRIC password.
PG is database handle."
  (let* ((value (sql-escape-string value))
	 (sql (format "UPDATE %s SET %s = pgp_sym_encrypt(%s,'%s') WHERE %s_id = %d" table column value symmetric table id)))
    (rcd-sql-first sql pg)))

(defun rcd-db-split-edit-id (id)
  (let* ((id (split-string id ":"))
	 (table (car id))
	 (column (cadr id))
	 (id (string-to-number (caddr id))))
    (list table column id)))

;;; TABLE MANAGEMENT

(defun rcd-db-create-table (table pg)
  "Create TABLE with database handle PG."
  (let* ((table (or table (rcd-ask-get "New database table name: ")))
	 (sql (format "CREATE TABLE %s (
                      %s_id SERIAL NOT NULL PRIMARY KEY,
                      %s_datecreated TIMESTAMP DEFAULT CURRENT_TIMESTAMP NOT NULL,
                      %s_datemodified TIMESTAMP,
                      %s_usercreated TEXT NOT NULL DEFAULT current_user,
                      %s_usermodified TEXT NOT NULL DEFAULT current_user,
                      %s_name TEXT NOT NULL,
                      %s_description TEXT)"
		      table table table table table table table table)))
    (rcd-sql sql pg)))

(defun rcd-db-drop-table (table pg)
  "Drop database TABLE with database handle PG."
  (rcd-sql (format "DROP TABLE %s" table) pg))

(defun rcd-db-alter-table-rename-column (table old-column-name new-column-name pg)
  "Alter TABLE and rename column from OLD-COLUMN-NAME to NEW-COLUMN-NAME"
  (let* ((sql (format "ALTER TABLE %s RENAME COLUMN %s TO %s" table old-column-name new-column-name)))
    (rcd-sql sql pg)))

(defun rcd-db-alter-table-enable-row-level-security (table pg)
  "Alter TABLE and enable row level security."
  (let* ((sql (format "ALTER TABLE %s ENABLE ROW LEVEL SECURITY" table)))
    (rcd-sql sql pg)))

(defun rcd-db-alter-table-add-column (table column type references not-null default pg)
  "Alter TABLE and add new COLUMN.
REFERENCES carry the name of table to be referenced, in that case
TYPE is not considered. NOT-NULL may be TRUE or NIL. DEFAULT may
be provided."
  (let* ((not-null (if not-null "NOT NULL" ""))
	 (type (if references "" type))
	 (references (if references (concat "INT4 REFERENCES " references) ""))
	 (default (if (length> default 0) (concat "DEFAULT " default) ""))
	 (sql (format "ALTER TABLE %s ADD COLUMN %s %s %s %s %s;" table column type references not-null default)))
    (rcd-sql sql pg)))

(defun rcd-db-alter-table-drop-column (table column pg)
  "Alter TABLE drop COLUMN for database handle PG."
  (rcd-sql (format "ALTER TABLE %s DROP COLUMN %s" table column) pg))

(defun rcd-db-column-default (table column pg &optional schema)
  "Return column default for TABLE, COLUMN, PG.

SCHEMA is optional, by default \"public\"."
  (let ((schema (or schema "public")))
    (rcd-sql-first "SELECT column_default FROM information_schema.columns 
                    WHERE (table_schema, table_name, column_name) = ($1, $2, $3)"
		   pg schema table column)))

;;; Timestamp conversion

(defun rcd-db-update-timestamp-with-time-zone ()
  "Update all columns without time zone to timestamp with time zone."
  (interactive)
  (let ((tables (rcd-db-tables-list cf-db)))
    (while tables 
      (let* ((table (pop tables))
	     (columns (rcd-db-table-columns-list table cf-db)))
	(while columns
	  (let* ((column (pop columns))
		 (type (rcd-db-column-type table column cf-db)))
	    (when (string= type "timestamp without time zone")
	      (let ((prompt (format 
			     "Do you wish to update for table `%s' and column `%s' to timestamp with time zone? "
			     table column)))
		(when (or (string-match (rx (seq "_date" (or "created" "modified"))) column)
			  (y-or-n-p prompt))
		(let ((sql (format "ALTER TABLE %s 
                                    ALTER COLUMN %s
                                     TYPE TIMESTAMP WITH TIME ZONE"
				   table column)))
		  (rcd-message "%s" sql)
		  (rcd-sql sql cf-db)))))))))))

;; (rcd-db-update-timestamp-with-time-zone)

;;; DEFAULT TRIGGERS FOR TABLES

(defun rcd-db-trigger-exists (trigger pg)
  "Check if TRIGGER exists for connection PG."
  (let ((sql (format "SELECT trigger_name FROM information_schema.triggers
                      WHERE trigger_name = '%s'" trigger)))
    (rcd-sql-first sql pg)))

(defun rcd-db-trigger-date-modified (table pg)
  "Create trigger for modified date using TABLE and PG." ;; TODO we did not check ifbbb columns exist
  (let ((trigger (concat table "_moddatetime")))
    (unless (rcd-db-trigger-exists trigger pg)
      (let ((sql (format "CREATE TRIGGER %s
                          BEFORE UPDATE ON %s
                          FOR EACH ROW
                          EXECUTE PROCEDURE moddatetime(%s_datemodified)"
			 trigger table table)))
	(rcd-sql sql pg)))))

(defun rcd-db-trigger-user-modified (table pg)
  "Create trigger for user modified TABLE and PG." ;; TODO we did not check ifbbb columns exist
  (let ((trigger (concat "insert_username_" table)))
    (unless (rcd-db-trigger-exists trigger pg)
      (let ((sql (format "CREATE TRIGGER %s
                          BEFORE INSERT OR UPDATE ON %s
                          FOR EACH ROW
                          EXECUTE PROCEDURE insert_username(%s_usermodified)"
			 trigger table table)))
	(rcd-sql sql pg)))))

;;; COLUMN FUNCTIONS

(defun rcd-db-column-width (table column pg)
  "Return COLUMN width for TABLE using PG."
  (rcd-sql-first (format "SELECT MAX(LENGTH(CAST(%s AS text))) AS max FROM %s" column table) pg))

;;; FANCY SQL FUNCTIONS

(defun rcd-sql-age (iso-date pg)
  "Return age from PostgreSQL by providing ISO-DATE and PG database handle."
  (let* ((sql (format "SELECT age(date('%s'))" iso-date))
	 (age (rcd-sql-first sql pg))
	 (age (replace-regexp-in-string "mon" "month" age)))
    age))

;;; SQL SKELETONS

(define-skeleton cf-sql-table
    "Prepare the SQL table for Central Files database design."
  nil
  "
-- ------------------------------------------
-- ------------ Table " (setq table (skeleton-read "Table name: ")) "
-- ------------------------------------------
DROP SEQUENCE " table "_id_seq;

CREATE TABLE " table " (
" table "_id INTEGER GENERATED BY DEFAULT AS IDENTITY,
" table "_uuid UUID NOT NULL DEFAULT gen_random_uuid(),
" table "_datecreated TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP NOT NULL,
" table "_datemodified TIMESTAMP WITH TIME ZONE,
" table "_usercreated TEXT NOT NULL DEFAULT current_user,
" table "_usermodified TEXT NOT NULL DEFAULT current_user,
" table "_name TEXT,
" table "_description TEXT,
" table "_ TEXT
);
GRANT ALL ON " table " TO PUBLIC;
DROP VIEW " table "_combo;
CREATE OR REPLACE VIEW " table "_combo AS
SELECT " table "_id AS id,
" table "_name AS TEXT
FROM " table ";
GRANT SELECT ON " table "_combo TO PUBLIC;
COMMENT ON TABLE " table " IS '" (capitalize table) "';
COMMENT ON COLUMN " table "." table "_id IS 'ID';
COMMENT ON COLUMN " table "." table "_uuid IS 'UUID';
COMMENT ON COLUMN " table "." table "_datecreated IS 'Date created';
COMMENT ON COLUMN " table "." table "_datemodified IS 'Date modified';
COMMENT ON COLUMN " table "." table "_usercreated IS 'User created';
COMMENT ON COLUMN " table "." table "_usermodified IS 'User modified';
COMMENT ON COLUMN " table "." table "_name IS 'Name';
COMMENT ON COLUMN " table "." table "_description IS 'Description';
COMMENT ON COLUMN " table "." table "_IS '';

CREATE UNIQUE INDEX " table "_index ON " table " ( " table "_name );

-- INSERT INTO " table " (" table "_name) VALUES ('');
-- INSERT INTO meta_tables VALUES ('" table "', 'hide', '1');

-- Triggers
-- For Date Modified
CREATE TRIGGER " table "_moddatetime
BEFORE UPDATE ON " table "
FOR EACH ROW
EXECUTE PROCEDURE moddatetime(" table "_datemodified);

-- For User Modified
CREATE TRIGGER insert_username_" table "
BEFORE INSERT OR UPDATE ON " table "
FOR EACH ROW
EXECUTE PROCEDURE insert_username(" table "_usermodified);

")

;;; PACKAGE FOOTER

(provide 'rcd-db-basics)

;;; rcd-db-basics.el ends here

;; Local Variables:
;; rcd-vc-ask-for-revision: t
;; rcd-vc-ask-for-description: nil
;; outline-heading-alist: '((";;;; " . 2))
;; End:
